@@ -1,7 +1,7 @@
 /*
  * Global program option structure and the parsing function prototype.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -34,6 +34,7 @@ struct opts_s {           /* structure describing run-time options */
 	bool average_rate;             /* average rate counter flag */
 	bool bytes;                    /* bytes transferred flag */
 	bool bits;                     /* report transfer size in bits */
+	bool decimal_units;            /* decimal prefix flag */
 	bool bufpercent;               /* transfer buffer percentage flag */
 	size_t lastwritten;            /* show N bytes last written */
 	bool force;                    /* force-if-not-terminal flag */
@@ -53,11 +54,12 @@ struct opts_s {           /* structure describing run-time options */
 	bool stop_at_size;             /* set if we stop at "size" bytes */
 	bool sync_after_write;         /* set if we sync after every write */
 	bool direct_io;                /* set if O_DIRECT is to be used */
-	bool discard_input;            /* set to write nothing to stdout */
+	bool discard_input;            /* set to write nothing to output */
 	double interval;               /* interval between updates */
 	double delay_start;            /* delay before first display */
 	pid_t watch_pid;	       /* process to watch fds of */
 	int watch_fd;		       /* fd to watch */
+	/*@keep@*/ /*@null@*/ char *output; /* fd to write output to */
 	unsigned int average_rate_window; /* time window in seconds for average rate calculations */
 	unsigned int width;            /* screen width */
 	unsigned int height;           /* screen height */
