@@ -2,7 +2,7 @@
  * Main program entry point - read the command line options, then perform
  * the appropriate actions.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -12,7 +12,7 @@
 #include "pv.h"
 
 /* We do not set this because it breaks "dd" - see below. */
-/* #undef MAKE_STDOUT_NONBLOCKING */
+/* #undef MAKE_OUTPUT_NONBLOCKING */
 
 #include <stdio.h>
 #include <string.h>
@@ -49,8 +49,8 @@ int main(int argc, char **argv)
 	/* Parse the command line arguments. */
 	opts = opts_parse(argc >= 0 ? (unsigned int) argc : 0, argv);
 	if (NULL == opts) {
-		debug("%s: %d", "exiting with status", 64);
-		return 64;
+		debug("%s: %d", "exiting with status", PV_ERROREXIT_MEMORY);
+		return PV_ERROREXIT_MEMORY;
 	}
 
 	/* Early exit if necessary, such as with "-h". */
@@ -73,8 +73,8 @@ int main(int argc, char **argv)
 		 */
 		fprintf(stderr, "%s: %s: %s\n", opts->program_name, _("state allocation failed"), strerror(errno));
 		opts_free(opts);
-		debug("%s: %d", "exiting with status", 64);
-		return 64;
+		debug("%s: %d", "exiting with status", PV_ERROREXIT_MEMORY);
+		return PV_ERROREXIT_MEMORY;
 		/*@+mustfreefresh@ */
 	}
 
@@ -115,7 +115,7 @@ int main(int argc, char **argv)
 			fprintf(stderr, "%s: %s\n", opts->program_name, strerror(errno));
 			pv_state_free(state);
 			opts_free(opts);
-			return 1;
+			return PV_ERROREXIT_REMOTE_OR_PID;
 		}
 		memset(pidfile_tmp_name, 0, pidfile_tmp_bufsize);
 		(void) pv_snprintf(pidfile_tmp_name, pidfile_tmp_bufsize, "%s.XXXXXX", opts->pidfile);
@@ -133,7 +133,7 @@ int main(int argc, char **argv)
 			free(pidfile_tmp_name);
 			pv_state_free(state);
 			opts_free(opts);
-			return 1;
+			return PV_ERROREXIT_REMOTE_OR_PID;
 		}
 
 		(void) umask(prev_umask);   /* flawfinder: ignore */
@@ -159,7 +159,7 @@ int main(int argc, char **argv)
 			free(pidfile_tmp_name);
 			pv_state_free(state);
 			opts_free(opts);
-			return 1;
+			return PV_ERROREXIT_REMOTE_OR_PID;
 		}
 
 		fprintf(pidfile_tmp_fptr, "%d\n", getpid());
@@ -183,7 +183,7 @@ int main(int argc, char **argv)
 		if (!opts_add_file(opts, "-")) {
 			pv_state_free(state);
 			opts_free(opts);
-			return 64;
+			return PV_ERROREXIT_MEMORY;
 		}
 	}
 
@@ -278,6 +278,7 @@ int main(int argc, char **argv)
 	pv_state_delay_start_set(state, opts->delay_start);
 	pv_state_linemode_set(state, opts->linemode);
 	pv_state_bits_set(state, opts->bits);
+	pv_state_decimal_units_set(state, opts->decimal_units);
 	pv_state_null_terminated_lines_set(state, opts->null_terminated_lines);
 	pv_state_skip_errors_set(state, opts->skip_errors);
 	pv_state_error_skip_block_set(state, opts->error_skip_block);
@@ -299,15 +300,38 @@ int main(int argc, char **argv)
 			    opts->fineta, opts->rate, opts->average_rate,
 			    opts->bytes, opts->bufpercent, opts->lastwritten, opts->name);
 
-#ifdef MAKE_STDOUT_NONBLOCKING
 	/*
-	 * Try and make standard output use non-blocking I/O.
+	 * Set output file, treating no output or "-" as stdout
+	 */
+	if (NULL == opts->output || 0 == strcmp(opts->output, "-")) {
+		pv_state_output_set(state, STDOUT_FILENO, "(stdout)");
+	} else {
+		int fd = open(opts->output, O_WRONLY | O_CREAT | O_TRUNC, 0600);	/* flawfinder: ignore */
+		/*
+		 * flawfinder rationale: the output filename has been
+		 * explicitly provided, and in many cases the operator will
+		 * want to write to device files and other special
+		 * destinations, so there is no sense-checking we can do to
+		 * make this safer.
+		 */
+		if (fd < 0) {
+			fprintf(stderr, "%s: %s: %s\n", opts->program_name, opts->output, strerror(errno));
+			pv_state_free(state);
+			opts_free(opts);
+			return PV_ERROREXIT_ACCESS;
+		}
+		pv_state_output_set(state, fd, opts->output);
+	}
+
+#ifdef MAKE_OUTPUT_NONBLOCKING
+	/*
+	 * Try and make the output use non-blocking I/O.
 	 *
 	 * Note that this can cause problems with (broken) applications
-	 * such as dd.
+	 * such as dd when used in a pipeline.
 	 */
-	fcntl(STDOUT_FILENO, F_SETFL, O_NONBLOCK | fcntl(STDOUT_FILENO, F_GETFL));
-#endif				/* MAKE_STDOUT_NONBLOCKING */
+	fcntl(state->control.output_fd, F_SETFL, O_NONBLOCK | fcntl(state->control.output_fd, F_GETFL));
+#endif				/* MAKE_OUTPUT_NONBLOCKING */
 
 	/* Initialise the signal handling. */
 	pv_sig_init(state);
