@@ -1,7 +1,7 @@
 /*
  * Functions providing the main transfer or file descriptor watching loop.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -192,6 +192,7 @@ int pv_main_loop(pvstate_t state)
 				&& (0 == state->control.rate_limit))) {
 				cansend = state->control.size - total_written;
 				if (0 >= cansend) {
+					debug("%s", "write limit reached (size explicitly set) - setting EOF flags");
 					eof_in = true;
 					eof_out = true;
 				}
@@ -207,6 +208,7 @@ int pv_main_loop(pvstate_t state)
 
 		/* End on write error. */
 		if (written < 0) {
+			debug("%s: %s", "write error from pv_transfer", strerror(errno));
 			if (state->control.cursor)
 				pv_crs_fini(state);
 			return state->status.exit_status;
@@ -344,6 +346,9 @@ int pv_main_loop(pvstate_t state)
 		transferred_since_last = 0;
 	}
 
+	debug("%s: %s=%s, %s=%s", "loop ended", "eof_in", eof_in ? "true" : "false", "eof_out",
+	      eof_out ? "true" : "false");
+
 	if (state->control.cursor) {
 		pv_crs_fini(state);
 	} else {
@@ -353,7 +358,7 @@ int pv_main_loop(pvstate_t state)
 	}
 
 	if (1 == state->flag.trigger_exit)
-		state->status.exit_status |= 32;
+		state->status.exit_status |= PV_ERROREXIT_SIGNAL;
 
 	if (fd >= 0)
 		(void) close(fd);
@@ -384,7 +389,7 @@ int pv_watchfd_loop(pvstate_t state)
 	info.watch_fd = state->control.watch_fd;
 	rc = pv_watchfd_info(state, &info, false);
 	if (0 != rc) {
-		state->status.exit_status |= 2;
+		state->status.exit_status |= PV_ERROREXIT_ACCESS;
 		/*@-compdestroy@ */
 		return state->status.exit_status;
 		/*@+compdestroy@ */
@@ -515,7 +520,7 @@ int pv_watchfd_loop(pvstate_t state)
 		pv_write_retry(STDERR_FILENO, "\n", 1);
 
 	if (1 == state->flag.trigger_exit)
-		state->status.exit_status |= 32;
+		state->status.exit_status |= PV_ERROREXIT_SIGNAL;
 
 	/*
 	 * Free the state structure specific to this file descriptor. 
@@ -568,8 +573,8 @@ int pv_watchpid_loop(pvstate_t state)
 	 */
 	if (kill(state->control.watch_pid, 0) != 0) {
 		pv_error(state, "%s %u: %s", _("pid"), state->control.watch_pid, strerror(errno));
-		state->status.exit_status |= 2;
-		return 2;
+		state->status.exit_status |= PV_ERROREXIT_ACCESS;
+		return PV_ERROREXIT_ACCESS;
 	}
 
 	/*
@@ -619,10 +624,10 @@ int pv_watchpid_loop(pvstate_t state)
 		if (kill(state->control.watch_pid, 0) != 0) {
 			if (first_pass) {
 				pv_error(state, "%s %u: %s", _("pid"), state->control.watch_pid, strerror(errno));
-				state->status.exit_status |= 2;
+				state->status.exit_status |= PV_ERROREXIT_ACCESS;
 				if (NULL != info_array)
 					free(info_array);
-				return 2;
+				return PV_ERROREXIT_ACCESS;
 			}
 			break;
 		}
@@ -660,10 +665,10 @@ int pv_watchpid_loop(pvstate_t state)
 		if (rc != 0) {
 			if (first_pass) {
 				pv_error(state, "%s %u: %s", _("pid"), state->control.watch_pid, strerror(errno));
-				state->status.exit_status |= 2;
+				state->status.exit_status |= PV_ERROREXIT_ACCESS;
 				if (NULL != info_array)
 					free(info_array);
-				return 2;
+				return PV_ERROREXIT_ACCESS;
 			}
 			break;
 		}
