@@ -1,7 +1,7 @@
 /*
  * Functions for transferring data between file descriptors.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -156,8 +156,8 @@ static ssize_t pv__transfer_read_repeated(int fd, void *buf, size_t count)
 		elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
 
 		if (elapsed_seconds > TRANSFER_READ_TIMEOUT) {
-			debug("%s %d: %s (%lf %s)", "fd", fd,
-			      "stopping read - timer expired", elapsed_seconds, "sec elapsed");
+			debug("%s %d: %s (%f %s)", "fd", fd,
+			      "stopping read - timer expired", (double) elapsed_seconds, "sec elapsed");
 			return total_read;
 		}
 
@@ -231,8 +231,8 @@ static ssize_t pv__transfer_write_repeated(int fd, void *buf, size_t count, bool
 			if ((EINTR == errno) || (EAGAIN == errno)) {
 				/*
 				 * Interrupted by a signal - probably our
-				 * alarm() - so just return what we've
-				 * written so far.
+				 * alarm or interval timer - so just return
+				 * what we've written so far.
 				 */
 				return total_written;
 			} else {
@@ -259,16 +259,17 @@ static ssize_t pv__transfer_write_repeated(int fd, void *buf, size_t count, bool
 		elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
 
 		if (elapsed_seconds > TRANSFER_WRITE_TIMEOUT) {
-			debug("%s %d: %s (%lf %s)", "fd", fd,
-			      "stopping write - timer expired", elapsed_seconds, "sec elapsed");
+			debug("%s %d: %s (%f %s)", "fd", fd,
+			      "stopping write - timer expired", (double) elapsed_seconds, "sec elapsed");
 			return total_written;
 		}
 
 		/*
 		 * Running the select() here seems to make PV eat a lot of
 		 * CPU in some cases, so instead we just go round the loop
-		 * again and rely on our alarm() to interrupt us if we run
-		 * out of time - also on our elapsed time check.
+		 * again and rely on our alarm or interval timer to
+		 * interrupt us if we run out of time - also on our elapsed
+		 * time check.
 		 */
 		if (count > 0) {
 			debug("%s %d: %s (%ld %s, %ld %s)", "fd", fd,
@@ -344,7 +345,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 		/*@-nullpass@ */
 		/*@-type@ */
 		/* splint doesn't know about splice */
-		nread = splice(fd, NULL, STDOUT_FILENO, NULL, bytes_to_splice, SPLICE_F_MORE);
+		nread = splice(fd, NULL, state->control.output_fd, NULL, bytes_to_splice, SPLICE_F_MORE);
 		/*@+type@ */
 		/*@+nullpass@ */
 
@@ -369,7 +370,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 				 * error, we cannot skip it, so set
 				 * "do_not_skip_errors".
 				 */
-				if ((fdatasync(STDOUT_FILENO) < 0)
+				if ((fdatasync(state->control.output_fd) < 0)
 				    && (EIO == errno)) {
 					nread = -1;
 					do_not_skip_errors = true;
@@ -445,7 +446,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 	 * exit status, regardless of whether we're skipping errors, and
 	 * increment the error counter.
 	 */
-	state->status.exit_status |= 16;
+	state->status.exit_status |= PV_ERROREXIT_TRANSFER;
 	state->transfer.read_errors_in_a_row++;
 
 	/*
@@ -611,13 +612,13 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 
 
 /*
- * Write state->transfer.to_write bytes of data from the transfer buffer to stdout.
+ * Write state->transfer.to_write bytes of data from the transfer buffer to the output.
  * Returns zero if there was a transient error and we need to return 0 from
  * pv_transfer, otherwise returns 1.
  *
  * Updates state->transfer.write_position by moving it on by the number of bytes
  * written; adds the number of bytes written to state->transfer.written; sets
- * *eof_out to true, on stdout EOF, or when the write position catches up
+ * *eof_out to true, on output EOF, or when the write position catches up
  * with the read position AND *eof_in is true (meaning we've reached the end
  * of data).
  *
@@ -632,7 +633,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 
 	if (NULL == state->transfer.transfer_buffer) {
 		pv_error(state, "%s", _("no transfer buffer allocated"));
-		state->status.exit_status |= 64;
+		state->status.exit_status |= PV_ERROREXIT_MEMORY;
 		*eof_out = true;
 		state->transfer.written = -1;
 		return 1;
@@ -643,22 +644,69 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 	if (state->control.discard_input) {
 		nwritten = state->transfer.to_write;
 	} else if (state->transfer.to_write > 0) {
-		if (signal(SIGALRM, SIG_IGN) == SIG_ERR) {
-			pv_error(state, "%s: %s", _("failed to set alarm signal handler"), strerror(errno));
-		} else {
-			(void) alarm(1);
+
+		/*
+		 * Set an interval timer or an alarm to interrupt the write
+		 * with a signal if the write takes too long, so we can
+		 * continue producing progress information.
+		 */
+#if HAVE_SETITIMER
+		struct itimerval new_timer;
+
+		/*@-unrecog@ */
+		/* splint doesn't know setitimer or ITIMER_REAL */
+		memset(&new_timer, 0, sizeof(new_timer));
+		new_timer.it_value.tv_sec = (time_t) (state->control.interval);
+		new_timer.it_value.tv_usec = (suseconds_t) (((long) (state->control.interval * 1000000.0)) % 1000000);
+
+		/*
+		 * We have to set the interval so that the timer continues
+		 * to repeat while writes are attempted, especially as it's
+		 * possible that the initial timer run will expire
+		 * immediately if the period is less than 1 second.
+		 */
+
+		new_timer.it_interval.tv_sec = new_timer.it_value.tv_sec;
+		new_timer.it_interval.tv_usec = new_timer.it_value.tv_usec;
+
+		debug("%s: [%lds,%ldus]", "setting interval timer", (long) (new_timer.it_value.tv_sec),
+		      (long) (new_timer.it_value.tv_usec));
+
+		if (0 != setitimer(ITIMER_REAL, &new_timer, NULL)) {
+			pv_error(state, "%s: %s", _("failed to set interval timer"), strerror(errno));
 		}
-		nwritten = pv__transfer_write_repeated(STDOUT_FILENO,
+
+#else				/* ! HAVE_SETITIMER */
+		(void) alarm(1);
+		debug("%s", "setting alarm");
+#endif				/* HAVE_SETITIMER */
+		debug("%s: %ld %s", "beginning write attempt", (long) (state->transfer.to_write), "bytes");
+		nwritten = pv__transfer_write_repeated(state->control.output_fd,
 						       state->transfer.transfer_buffer +
 						       state->transfer.write_position,
 						       (size_t) (state->transfer.to_write),
 						       state->control.sync_after_write);
+		debug("%s: %ld", "bytes written", (long) nwritten);
+#if HAVE_SETITIMER
+		memset(&new_timer, 0, sizeof(new_timer));
+		new_timer.it_interval.tv_sec = 0;
+		new_timer.it_interval.tv_usec = 0;
+		new_timer.it_value.tv_sec = 0;
+		new_timer.it_value.tv_usec = 0;
+		if (0 != setitimer(ITIMER_REAL, &new_timer, NULL)) {
+			pv_error(state, "%s: %s", _("failed to clear interval timer"), strerror(errno));
+		}
+
+		/*@+unrecog@ */
+#else				/* ! HAVE_SETITIMER */
+		debug("%s", "cancelling alarm");
 		(void) alarm(0);
+#endif				/* HAVE_SETITIMER */
 	}
 
 	if (0 == nwritten) {
 		/*
-		 * Write returned 0 - EOF on stdout.
+		 * Write returned 0 - EOF on output.
 		 */
 		*eof_out = true;
 		return 1;
@@ -772,7 +820,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 	}
 
 	pv_error(state, "%s: %s", _("write failed"), strerror(errno));
-	state->status.exit_status |= 16;
+	state->status.exit_status |= PV_ERROREXIT_TRANSFER;
 	*eof_out = true;
 	state->transfer.written = -1;
 
@@ -794,7 +842,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
  */
 /*@null@*/
 /*@only@*/
-static char *pv__allocate_aligned_buffer(int fd, size_t target_size)
+static char *pv__allocate_aligned_buffer(int outfd, int infd, size_t target_size)
 {
 	char *newptr;
 
@@ -802,8 +850,8 @@ static char *pv__allocate_aligned_buffer(int fd, size_t target_size)
 	long input_alignment, output_alignment, min_alignment;
 	long required_alignment;
 
-	input_alignment = fd >= 0 ? fpathconf(fd, _PC_REC_XFER_ALIGN) : -1;
-	output_alignment = fpathconf(STDOUT_FILENO, _PC_REC_XFER_ALIGN);
+	input_alignment = infd >= 0 ? fpathconf(infd, _PC_REC_XFER_ALIGN) : -1;
+	output_alignment = fpathconf(outfd, _PC_REC_XFER_ALIGN);
 #if defined(HAVE_SYSCONF) && defined(_SC_PAGESIZE)
 	min_alignment = sysconf(_SC_PAGESIZE);
 #else				/* ! defined(HAVE_SYSCONF) && defined(_SC_PAGESIZE) */
@@ -880,9 +928,12 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 			}
 		}
 		if (!(*eof_out)) {
-			if (0 != fcntl(STDOUT_FILENO, F_SETFL,
-				       (state->control.direct_io ? O_DIRECT : 0) | fcntl(STDOUT_FILENO, F_GETFL))) {
-				debug("%s: %s: %s", "(stdout)", "fcntl", strerror(errno));
+			if (0 != fcntl(state->control.output_fd, F_SETFL,
+				       (state->control.direct_io ? O_DIRECT : 0) |
+				       fcntl(state->control.output_fd, F_GETFL))) {
+				debug("%s: %s: %s",
+				      NULL == state->control.output_name ? "(null)" : state->control.output_name,
+				      "fcntl", strerror(errno));
 			}
 		}
 		state->control.direct_io_changed = false;
@@ -905,10 +956,10 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	 */
 	if (NULL == state->transfer.transfer_buffer) {
 		state->transfer.transfer_buffer =
-		    pv__allocate_aligned_buffer(fd, state->control.target_buffer_size + 32);
+		    pv__allocate_aligned_buffer(state->control.output_fd, fd, state->control.target_buffer_size + 32);
 		if (NULL == state->transfer.transfer_buffer) {
 			pv_error(state, "%s: %s", _("buffer allocation failed"), strerror(errno));
-			state->status.exit_status |= 64;
+			state->status.exit_status |= PV_ERROREXIT_MEMORY;
 			return -1;
 		}
 		state->transfer.buffer_size = state->control.target_buffer_size;
@@ -923,7 +974,8 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	 */
 	if (state->transfer.buffer_size < state->control.target_buffer_size) {
 		char *newptr;
-		newptr = pv__allocate_aligned_buffer(fd, state->control.target_buffer_size + 32);
+		newptr =
+		    pv__allocate_aligned_buffer(state->control.output_fd, fd, state->control.target_buffer_size + 32);
 		if (NULL == newptr) {
 			/*
 			 * Reset target if realloc failed so we don't keep
@@ -983,10 +1035,10 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 
 	/*
 	 * If we don't think we've finished writing and there's anything
-	 * we're allowed to write, look for the stdout becoming writable.
+	 * we're allowed to write, look for the output becoming writable.
 	 */
 	if ((!(*eof_out)) && (state->transfer.to_write > 0)) {
-		check_write_fd = STDOUT_FILENO;
+		check_write_fd = state->control.output_fd;
 	}
 
 	ready_to_read = false;
@@ -1009,7 +1061,7 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 		/*@+compdef@ */
 		/* splint - see previous pv_current_file_name() calls. */
 
-		state->status.exit_status |= 16;
+		state->status.exit_status |= PV_ERROREXIT_TRANSFER;
 
 		return -1;
 	}
@@ -1044,9 +1096,9 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	}
 
 	/*
-	 * If there is data to write, and stdout is ready to receive it, and
-	 * we didn't use splice() this time, write some data.  Return early
-	 * if there was a transient write error.
+	 * If there is data to write, and the output is ready to receive it,
+	 * and we didn't use splice() this time, write some data. 
+	 * Return early if there was a transient write error.
 	 */
 	if (ready_to_write
 #ifdef HAVE_SPLICE
